/*****************************************************************************\
 *                        ANALYSIS PERFORMANCE TOOLS                         *
 *                                   Extrae                                  *
 *              Instrumentation package for parallel applications            *
 *****************************************************************************
 *     ___     This library is free software; you can redistribute it and/or *
 *    /  __         modify it under the terms of the GNU LGPL as published   *
 *   /  /  _____    by the Free Software Foundation; either version 2.1      *
 *  /  /  /     \   of the License, or (at your option) any later version.   *
 * (  (  ( B S C )                                                           *
 *  \  \  \_____/   This library is distributed in hope that it will be      *
 *   \  \__         useful but WITHOUT ANY WARRANTY; without even the        *
 *    \___          implied warranty of MERCHANTABILITY or FITNESS FOR A     *
 *                  PARTICULAR PURPOSE. See the GNU LGPL for more details.   *
 *                                                                           *
 * You should have received a copy of the GNU Lesser General Public License  *
 * along with this library; if not, write to the Free Software Foundation,   *
 * Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA          *
 * The GNU LEsser General Public License is contained in the file COPYING.   *
 *                                 ---------                                 *
 *   Barcelona Supercomputing Center - Centro Nacional de Supercomputacion   *
\*****************************************************************************/

#include "common.h"

#ifdef HAVE_UNISTD_H
# include <unistd.h>
#endif
#ifdef HAVE_STDLIB_H
# include <stdlib.h>
#endif
#ifdef HAVE_STDIO_H
# ifdef HAVE_FOPEN64
#  define __USE_LARGEFILE64
# endif
# include <stdio.h>
#endif
#ifdef HAVE_ERRNO_H
# include <errno.h>
#endif
#ifdef HAVE_TIME_H
# include <time.h>
#endif
#ifdef HAVE_SYS_TYPES_H
# include <sys/types.h>
#endif
#ifdef HAVE_SYS_STAT_H
# include <sys/stat.h>
#endif
#ifdef HAVE_FCNTL_H
# include <fcntl.h>
#endif
#ifdef HAVE_STRING_H
# include <string.h>
#endif
#ifdef HAVE_SYS_TIME_H
# include <sys/time.h>
#endif
#ifdef HAVE_LIBGEN_H
# include <libgen.h>
#endif
#if defined (PARALLEL_MERGE)
# include <mpi.h>
# include "mpi-tags.h"
# include "mpi-aux.h"
# include "parallel_merge_aux.h"
#endif

#include "utils.h"
#include "file_set.h"
#include "events.h"
#include "record.h"
#include "object_tree.h"
#include "mpi2out.h"
#include "trace_to_prv.h"
#include "mpi_prv_semantics.h"
#include "misc_prv_semantics.h"
#include "omp_prv_semantics.h"
#include "pthread_prv_semantics.h"
#include "gaspi_prv_semantics.h"
#include "mpi_comunicadors.h"
#include "labels.h"
#include "trace_mode.h"
#include "semantics.h"
#include "dump.h"
#include "communicators.h"
#include "cpunode.h"
#include "checkoptions.h"
#include "options.h"

#include "paraver_state.h"
#include "paraver_generator.h"
#include "dimemas_generator.h"

#include "mpi_prv_events.h"
#include "pthread_prv_events.h"
#include "openshmem_prv_events.h"
#include "openacc_prv_events.h"
#include "omp_prv_events.h"
#include "misc_prv_events.h"
#include "opencl_prv_events.h"
#include "cuda_prv_events.h"
#include "hip_prv_events.h"
#include "java_prv_events.h"
#include "gaspi_prv_events.h"
#include "addr2info.h"
#include "timesync.h"
#include "vector.h"
#include "xalloc.h"

#if USE_HARDWARE_COUNTERS
# include "HardwareCounters.h"
#endif

int **EnabledTasks = NULL;
unsigned long long **EnabledTasks_time = NULL;
struct address_collector_t CollectedAddresses;

mpi2prv_vector_t *RegisteredStackValues = NULL;
Extrae_Vector_t RegisteredCodeLocationTypes;

static void InitializeEnabledTasks (int numberoftasks, int numberofapplications)
{
  int i, j;

  EnabledTasks = (int **) xmalloc (sizeof (int *) * numberofapplications);
  for (i = 0; i < numberofapplications; i++)
  {
    EnabledTasks[i] = (int *) xmalloc (sizeof (int) * numberoftasks);
    for (j = 0; j < numberoftasks; j++)
      EnabledTasks[i][j] = TRUE;
  }

  EnabledTasks_time =
    (unsigned long long **) xmalloc (sizeof (unsigned long long *) *
                                    numberofapplications);
  for (i = 0; i < numberofapplications; i++)
  {
    EnabledTasks_time[i] =
      (unsigned long long *) xmalloc (sizeof (unsigned long long) *
                                     numberoftasks);
    for (j = 0; j < numberoftasks; j++)
      EnabledTasks_time[i][j] = 0LL;
  }
}

static void HandleStackedType (unsigned ptask, unsigned task, unsigned thread,
	unsigned EvType, event_t *current_event)
{
	if (Get_EvEvent(current_event) == USER_EV &&
	    Vector_Search (RegisteredStackValues, EvType))
	{
		unsigned pos;
		unsigned u;
		int found;
		task_t *task_info = ObjectTree_getTaskInfo(ptask, task);
		thread_t *thread_info = ObjectTree_getThreadInfo(ptask, task, thread);
		active_task_thread_t *att = &(task_info->active_task_threads[thread_info->active_task_thread-1]);

		/* Look for existing stacked_types for the current task/thread */
		for (found = FALSE, u = 0; u < att->num_stacks && !found; u++)
			if (att->stacked_type[u].type == EvType)
			{
				pos = u;
				found = TRUE;
			}
		if (!found)
		{
			/* If wasn't find, this is the first event for such type in
			   this task/thread, create it */
			pos = att->num_stacks;

			att->stacked_type = (active_task_thread_stack_type_t*) xrealloc
			  (att->stacked_type, sizeof(active_task_thread_stack_type_t)*(pos+1));
			att->stacked_type[pos].stack = Stack_Init();
			att->stacked_type[pos].type = EvType;
			att->num_stacks++;
		}

		if (Get_EvMiscParam(current_event) > 0)
			Stack_Push (att->stacked_type[pos].stack, Get_EvMiscParam(current_event));
		else if (Get_EvMiscParam (current_event) == 0)
			Stack_Pop (att->stacked_type[pos].stack);
	}
}


/******************************************************************************
 ***  Paraver_ProcessTraceFiles
 ******************************************************************************/

int Paraver_ProcessTraceFiles (unsigned long nfiles,
	struct input_t *files, unsigned int num_appl,
	struct Pair_NodeCPU *NodeCPUinfo, int numtasks, int taskid)
{
	struct timeval time_begin, time_end;
	FileSet_t * fset;
	unsigned int cpu, ptask, task, thread, error;
	event_t * current_event;
	char envName[PATH_MAX], *tmp;
	unsigned int Type, EvType;
	unsigned long long current_time = 0;
	unsigned long long num_of_events, parsed_events, tmp_nevents;
	unsigned long long records_per_task;
	double pct, last_pct;
	UINT64 *StartingTimes, *SynchronizationTimes;
	unsigned i;
	long long options;
	int num_appl_tasks[num_appl];

	records_per_task = 1024*1024/sizeof(paraver_rec_t); /* num of events in 1 Mbytes */
	records_per_task *= get_option_merge_MaxMem();            /* let's use this memory */
#if defined(PARALLEL_MERGE)
	records_per_task /= get_option_merge_TreeFanOut();        /* divide by the tree fan out */
#endif

	ObjectTree_Initialize (taskid, num_appl, files, nfiles);
	for (i = 0; i < num_appl; i++)
		num_appl_tasks[i] = (ObjectTree_getPtaskInfo(i+1))->ntasks;

#if defined(PARALLEL_MERGE)
	ParallelMerge_InitCommunicators();
	InitPendingCommunication ();
	InitForeignRecvs (numtasks);
#endif
	Semantics_Initialize (PRV_SEMANTICS);

	fset = Create_FS (nfiles, files, taskid, PRV_SEMANTICS);
	error = (fset == NULL);

	Labels_loadLocalSymbols (taskid, nfiles, files, &StartingTimes, &SynchronizationTimes);
 	/* Generate a unique filename (use default or add suffix if needed) */
	if (taskid == 0){
		Extrae_GenerateOutputFileName();
	}

#if defined(PARALLEL_MERGE)
	if (numtasks > 1)
	{
		int res;
		unsigned int temp;

		res = MPI_Allreduce (&error, &temp, 1, MPI_UNSIGNED, MPI_LOR, MPI_COMM_WORLD);
		MPI_CHECK(res, MPI_Allreduce, "Failed to share translation completion!");

		error = temp;
	}
#endif

	if (error)
	{
		if (0 == taskid)
		{
			fprintf (stderr, "mpi2prv: Error! Some of the processors failed create trace descriptors\n");
			fflush (stderr);
		}
		return -1;
	}

	if (get_option_merge_Dump())
		make_dump (fset);

#if defined(HETEROGENEOUS_SUPPORT)
	/*	Must we accomodate all the intermediate files into this machine endianes? */
	EndianCorrection (fset, numtasks, taskid);
#endif

	/*
	 * Initialize the communicators management
	 */
	initialize_comunicadors (num_appl);

	options = GetTraceOptions (fset, numtasks, taskid);

	/* CheckHWCcontrol (taskid, options); */
	CheckClockType (taskid, options, PRV_SEMANTICS, get_option_merge_ForceFormat());

/**************************************************************************************/

#if defined(DEBUG)
	for (i = 0; i < nfiles; i++)
	{
		fprintf (stderr, "[DEBUG] SynchronizationTimes[%d] = %llu "
		                 " files[%d].SpawnOffset = %llu\n",
		                 i, SynchronizationTimes[i], i, files[i].SpawnOffset);
	}
#endif

	TimeSync_Initialize (num_appl, num_appl_tasks);
	for (i = 0; i < nfiles; i++)
		if (files[i].thread-1 == 0)
			TimeSync_SetInitialTime (files[i].ptask-1,
			  files[i].task-1,
			  StartingTimes[i],
			  SynchronizationTimes[i] - files[i].SpawnOffset,
			  files[i].node);

	if (get_option_merge_SincronitzaTasks_byNode())
	{
		if (0 == taskid)
			fprintf (stdout, "mpi2prv: Enabling Time Synchronization (Node).\n");
		TimeSync_CalculateLatencies (TS_NODE, get_option_merge_SincronitzaApps());
	}
	else if (get_option_merge_SincronitzaTasks())
	{
		if (0 == taskid)
			fprintf (stdout, "mpi2prv: Enabling Time Synchronization (Task).\n");
		TimeSync_CalculateLatencies (TS_TASK, get_option_merge_SincronitzaApps());
	}
	else 
	{
		if (0 == taskid)
			fprintf(stdout, "mpi2prv: Time Synchronization disabled.\n");
		TimeSync_CalculateLatencies (TS_NOSYNC, 0);
	}
	
/**************************************************************************************/

	CheckCircularBufferWhenTracing (fset, numtasks, taskid);

#if defined(MPI_PHYSICAL_COMM)
	BuscaComunicacionsFisiques (fset);
#endif

	InitializeEnabledTasks (nfiles, num_appl);

	error = FALSE;

	Initialize_States (fset);
	AddressCollector_Initialize (&CollectedAddresses);

	RegisteredStackValues = Vector_Init();
	Extrae_Vector_Init (&RegisteredCodeLocationTypes);

	if (0 == taskid)
	{
		fprintf (stdout, "mpi2prv: Parsing intermediate files\n");
		if (1 == numtasks)
			fprintf (stdout, "mpi2prv: Progress 1 of 2 ... ");
		fflush (stdout);
	}

	Rewind_FS (fset);

	parsed_events = 0;
	last_pct = 0.0;
	num_of_events = EventsInFS (fset);

	current_event = GetNextEvent_FS (fset, &cpu, &ptask, &task, &thread);

	if (taskid == 0)
	{
		gettimeofday (&time_begin, NULL);
	}

	do
	{
		tmp_nevents = 0;

		EvType = Get_EvEvent (current_event);

		if (getEventType (EvType, &Type))
		{
			current_time = TIMESYNC(ptask-1, task-1, Get_EvTime (current_event));

#if defined(DEBUG)
			fprintf (stderr, "mpi2prv: Parsing event %u:%u:%u <%u,%llu(%llu)> @%llu|%llu\n", ptask, task, thread, Get_EvEvent (current_event), Get_EvValue (current_event), Get_EvMiscParam(current_event), current_time, Get_EvTime(current_event));
#endif

			if (Type == PTHREAD_TYPE || Type == OPENMP_TYPE || Type == MISC_TYPE ||
			    Type == MPI_TYPE || Type == CUDA_TYPE || Type == HIP_TYPE || Type == OPENCL_TYPE ||
			    Type == OPENSHMEM_TYPE || Type == JAVA_TYPE || Type == OPENACC_TYPE ||
				Type == GASPI_TYPE)
			{
				task_t *task_info = ObjectTree_getTaskInfo(ptask, task);
				Ev_Handler_t *handler = Semantics_getEventHandler (EvType);
				if (handler != NULL)
				{
					handler (current_event, current_time, cpu, ptask, task, thread, fset);
					tmp_nevents = 1;

					if (MISC_TYPE == Type)
						Enable_MISC_Operation (EvType);
					else if (PTHREAD_TYPE == Type)
						Enable_pthread_Operation (EvType);
					else if (OPENMP_TYPE == Type)
						Enable_OMP_Operation (EvType);
					else if (MPI_TYPE == Type)
						Enable_MPI_Operation (EvType);
					else if (CUDA_TYPE == Type)
						Enable_CUDA_Operation (EvType, Get_EvValue(current_event));
					else if (HIP_TYPE == Type)
						Enable_HIP_Operation (EvType, Get_EvValue(current_event));
					else if (OPENCL_TYPE == Type)
						Enable_OpenCL_Operation (EvType);
					else if (OPENSHMEM_TYPE == Type)
						Enable_OPENSHMEM_Operation (EvType);
					else if (JAVA_TYPE == Type)
						Enable_Java_Operation (EvType);
					else if (OPENACC_TYPE == Type)
						Enable_OPENACC_Operation(EvType);
					else if (GASPI_TYPE == Type)
						Enable_GASPI_Operation(EvType, Get_EvParam(current_event));
				}
				else
				{
					fprintf (stderr, "mpi2prv: Error! unregistered event type %d in %s\n", EvType, __func__);
				}

				/* Deal with Nanos Task View if we have registered stacked values
				   and if we have seen a resume/suspend thread */
				if (!get_option_merge_NanosTaskView() &&
				    Vector_Count(RegisteredStackValues) > 0 &&
				    task_info->num_active_task_threads > 0)
				{
					int found;
					unsigned u;
					Extrae_Addr2Type_t *addr2types = NULL;

					HandleStackedType (ptask, task, thread, Get_EvValue(current_event), current_event);

					for (found = FALSE, u = 0; u < Extrae_Vector_Count (&RegisteredCodeLocationTypes) && !found; u++)
					{
						addr2types = Extrae_Vector_Get (&RegisteredCodeLocationTypes, u);
						found = addr2types->LineType == Get_EvValue(current_event);
					}
					if (found)
						HandleStackedType (ptask, task, thread, addr2types->FunctionType, current_event);
				}

#if USE_HARDWARE_COUNTERS || defined(HETEROGENEOUS_SUPPORT)
				if (Get_EvHWCRead (current_event))
				{
					thread_t *Sthread = ObjectTree_getThreadInfo(ptask, task, thread);
 
					if (Sthread->last_hw_group_change != current_time)
					{
						int i;
						unsigned int hwctype[MAX_HWC];
						unsigned long long hwcvalue[MAX_HWC];
						int num_hwc = 0;

						num_hwc = HardwareCounters_Emit (ptask, task, thread, current_time, current_event, hwctype, hwcvalue, FALSE);
						for (i = 0; i < num_hwc; i++)
							trace_paraver_event (cpu, ptask, task, thread, current_time, hwctype[i], hwcvalue[i]);

						if (get_option_merge_AbsoluteCounters())
						{
							num_hwc = HardwareCounters_Emit (ptask, task, thread, current_time, current_event, hwctype, hwcvalue, TRUE);
							for (i = 0; i < num_hwc; i++)
								trace_paraver_event (cpu, ptask, task, thread, current_time, hwctype[i], hwcvalue[i]);
						}
					}
				}
#endif
			}
			else if (Type == MPI_COMM_ALIAS_TYPE)
			{
				error = GenerateAliesComunicator (current_event, current_time, cpu, ptask, task, thread, fset, &tmp_nevents, PRV_SEMANTICS);
				Enable_MPI_Operation (EvType);
			}
		}
		else
		{
			tmp_nevents = 1;
		}

		(ObjectTree_getThreadInfo(ptask,task,thread))->First_Event = FALSE;
		(ObjectTree_getThreadInfo(ptask,task,thread))->Previous_Event_Time = current_time;

		parsed_events += tmp_nevents;
		pct = ((double) parsed_events)/((double) num_of_events)*100.0f;
		/* Right now, progress bar is only shown when numtasks is 1 */
		if (1 == numtasks)
		{
			if (pct > last_pct + 5.0 && pct <= 100.0f)
			{
				fprintf (stdout, "%d%% ", (int) pct);
				fflush (stdout);
				while (last_pct + 5.0 < pct)
					last_pct += 5.0;
			}
		}

		current_event = GetNextEvent_FS (fset, &cpu, &ptask, &task, &thread);

		long maxpct = get_option_merge_StopAtPercentage();
		if ((maxpct > 0) && (maxpct < 100) && (pct >= maxpct))
			current_event = NULL;

	} while ((current_event != NULL) && !error);

	if (1 == numtasks)
	{
		fprintf(stdout, "done\n");
		fflush(stdout);
	}

	fprintf(stdout, "mpi2prv: Processor %d %s to translate its assigned files\n", taskid, error?"failed":"succeeded");
	fflush(stdout);

#if defined(PARALLEL_MERGE)
	if (numtasks > 1)
	{
		int res, i;
		unsigned int temp, max_param;
		UINT64 temp2;

		res = MPI_Allreduce (&error, &temp, 1, MPI_UNSIGNED, MPI_LOR, MPI_COMM_WORLD);
		MPI_CHECK(res, MPI_Allreduce, "Failed to share translation completion!");
		error = temp;

		if (!error)
		{
			res = MPI_Allreduce (&current_time, &temp2, 1, MPI_LONG_LONG, MPI_MAX, MPI_COMM_WORLD);
			MPI_CHECK(res, MPI_Allreduce, "Failed to share end time!");
			current_time = temp2;

			for (i = 0; i<MAX_GASPI_PARAM_TYPE_ENTRIES; i++)
			{
				res = MPI_Allreduce(&GASPI_param_type_label[i].present, &max_param, 1, MPI_UNSIGNED, MPI_MAX, MPI_COMM_WORLD);
				MPI_CHECK(res, MPI_Allreduce, "Failed to share GASPI parameters!");
				GASPI_param_type_label[i].present = max_param;
			}
		}
	}
#endif

	if (taskid == 0)
	{
		time_t delta;
		gettimeofday (&time_end, NULL);
		delta = time_end.tv_sec - time_begin.tv_sec;
		fprintf (stdout, "mpi2prv: Elapsed time translating files: %ld hours %ld minutes %ld seconds\n", delta / 3600, (delta % 3600)/60, (delta % 60));
	}

	if (error)
	{
		if (0 == taskid)
		{
			fprintf (stderr, "mpi2prv: Error! Some of the processors failed to translate its files\n");
			fflush (stderr);
		}
		return -1;
	}

	/* Finalize states */
	Finalize_States (fset, current_time);

	/* Dump the Write File Buffer structure. As Finalize_States creates a new
	   paraver_rec_t for a new state, remove it (so TRUE in 2nd param) */
	Flush_FS (fset, FALSE);

	gettimeofday (&time_begin, NULL);
	Address2Info_Unify(numtasks, taskid, &CollectedAddresses);
	if (taskid == 0)
	{
		time_t delta;
		gettimeofday (&time_end, NULL);
		delta = time_end.tv_sec - time_begin.tv_sec;
		fprintf (stdout, "mpi2prv: Elapsed time unifying addresses: %ld hours %ld minutes %ld seconds\n", delta / 3600, (delta % 3600)/60, (delta % 60));
	}

#if defined(PARALLEL_MERGE)
	if (taskid == 0) gettimeofday (&time_begin, NULL);

	ParallelMerge_BuildCommunicators (numtasks, taskid);

	/* In the parallel merge we have to */
	if (numtasks > 1)
	{
		NewDistributePendingComms (numtasks, taskid, get_option_merge_UseDiskForComms());
		ShareTraceInformation (numtasks, taskid);
	}

	if (taskid == 0)
	{
		time_t delta;
		gettimeofday (&time_end, NULL);
		delta = time_end.tv_sec - time_begin.tv_sec;
		fprintf (stdout, "mpi2prv: Elapsed time sharing communications: %ld hours %ld minutes %ld seconds\n", delta / 3600, (delta % 3600)/60, (delta % 60));
	}
#endif

	error = Paraver_JoinFiles (num_appl, get_merge_OutputFileName (TRACE_FILENAME),
	  fset, current_time, NodeCPUinfo, numtasks,
	  taskid, records_per_task, get_option_merge_TreeFanOut());

	if (0 == taskid)
	{
		if (Labels_GeneratePCFfile (get_merge_OutputFileName (PCF_FILENAME), options) == -1)
			fprintf (stderr, "mpi2prv: WARNING! Unable to create PCF file!\n");

		if (GenerateROWfile (get_merge_OutputFileName (ROW_FILENAME), NodeCPUinfo, nfiles, files) == -1)
			fprintf (stderr, "mpi2prv: WARNING! Unable to create ROW file!\n");
	}

	if (0 == taskid)
	{
		if (error == 0)
		{
			fprintf (stdout, "mpi2prv: Congratulations! %s has been generated.\n",
			    get_merge_OutputFileName (TRACE_FILENAME));
		} else
		{
			fprintf (stdout, "mpi2prv: WARNING! Merge process finished with errors. Trace file may be incomplete.\n");
		}
		fflush (stdout);
	}

	return 0;
}
